@@ -1,214 +1,23 @@
-"""
-We want to parse a token stream into an AST, and then generate a new token stream from that AST.
-The generation should be a dual to the parsing. By 'dual' we mean that the generation algebra should be 
-as close as possible to the parsing algebra. The closest algebra to the parsing algebra is the parsing 
-algebra itself. 
-
-Given:
-AST = Syntax(Parser)(ParserState([Token, ...]))
-AST =?= Syntax(Parser)(GenState(AST))
-
-where =?= means the LHS and RHS induce the same text output, e.g. the same token stream 
-inspite of the token metadata, token types, and/or potentially different structure of the AST.
-
-With the above setting, Generator as a dual to Parser, can reuse most of the parsing combinator, the 
-change needed is to introduce randomness in the generation process, e.g. to generate a random variable name, etc.
-
-[Token, ...] == Syntax(Generator)(GenState(AST))
-
-"""
-
-
 from __future__ import annotations
-
 from typing import (
     Optional, List, Any, TypeVar, Generic, Callable, Tuple, cast, 
-    Dict, Type, ClassVar, Hashable,
-    Mapping, Iterator
+    Dict, Type, ClassVar, Hashable
 )
 
 import traceback
-from dataclasses import dataclass, fields, replace, field
-from functools import cached_property
+from dataclasses import dataclass, replace
 from weakref import WeakKeyDictionary
-from abc import ABC, abstractmethod
-from enum import Enum
+from abc import ABC
+from syncraft.ast import ThenKind, Then, Choice, Many, ChoiceKind, shallow_dict
+from syncraft.constraint import Bindable
 
 
 
 
-class Insptectable(ABC):
-    @abstractmethod
-    def to_string(self, interested: Callable[[Any], bool]) -> Optional[str]:
-        raise NotImplementedError("Subclasses must implement to_string")
+S = TypeVar('S', bound=Bindable)
     
-    @cached_property
-    def _string(self)->Optional[str]:
-        return self.to_string(lambda _: True)
-
-    def __repr__(self) -> str:
-        return self._string or self.__class__.__name__
-    def __str__(self) -> str:
-        return self._string or self.__class__.__name__
-
-
 A = TypeVar('A')  # Result type
-B = TypeVar('B')  # Result type for mapping
-
-S = TypeVar('S')  # State type for the Algebra
-
-    
-    
-class FrozenDict(Generic[A]):
-    def __init__(self, items: Mapping[str, A]):
-        for k, v in items.items():
-            if not isinstance(k, Hashable) or not isinstance(v, Hashable):
-                raise TypeError(f"Metadata key or value not hashable: {k} = {v}")
-        self._items = tuple(sorted(items.items()))
-
-    def __getitem__(self, key: str) -> A:
-        return dict(self._items)[key]
-
-    def __contains__(self, key: str) -> bool:
-        return key in dict(self._items)
-
-    def items(self) -> Iterator[tuple[str, A]]:
-        return iter(self._items)
-
-    def to_dict(self) -> dict[str, A]:
-        return dict(self._items)
-
-    def __hash__(self) -> int:
-        return hash(self._items)
-
-    def __eq__(self, other: object) -> bool:
-        return isinstance(other, FrozenDict) and self._items == other._items
-
-    def __repr__(self) -> str:
-        return f"FrozenDict({dict(self._items)})"
-
-
-
-@dataclass(frozen=True)
-class Lens(Generic[S, A]):
-    get: Callable[[S], A]
-    set: Callable[[S, A], S]    
-
-    def modify(self, source: S, f: Callable[[A], A]) -> S:
-        return self.set(source, f(self.get(source)))
-    
-    def bimap(self, ff: Callable[[A], B], bf: Callable[[B], A]) -> Lens[S, B]:
-        def getf(data: S) -> B:
-            return ff(self.get(data))
-
-        def setf(data: S, value: B) -> S:
-            return self.set(data, bf(value))
-
-        return Lens(get=getf, set=setf)
-
-    def __truediv__(self, other: Lens[A, B]) -> Lens[S, B]:
-        def get_composed(obj: S) -> B:
-            return other.get(self.get(obj))        
-        def set_composed(obj: S, value: B) -> S:
-            return self.set(obj, other.set(self.get(obj), value))
-        return Lens(get=get_composed, set=set_composed)
-    
-    def __rtruediv__(self, other: Lens[B, S])->Lens[B, A]:
-        return other.__truediv__(self)
-        
-class StructuralResult:
-    def bimap(self, ctx: Any)->Tuple[Any, Callable[[Any], StructuralResult]]:
-        return (self, lambda x: self)
-
-        
-@dataclass(frozen=True)
-class NamedResult(Generic[A], StructuralResult):
-    name: str
-    value: A
-    def bimap(self, ctx: Any)->Tuple[NamedResult[Any], Callable[[NamedResult[Any]], StructuralResult]]:
-        value, backward = self.value.bimap(ctx) if isinstance(self.value, StructuralResult) else (self.value, lambda x: x)
-        def named_back(data: Any)->NamedResult[Any]:
-            v = backward(data)
-            if isinstance(v, NamedResult):
-                return replace(v, name=self.name)
-            else:
-                return NamedResult(name=self.name, value=v)
-        return NamedResult(self.name, value), named_back
-
-@dataclass(eq=True, frozen=True)
-class ManyResult(Generic[A], StructuralResult):
-    value: Tuple[A, ...]
-    def bimap(self, ctx: Any)->Tuple[List[Any], Callable[[List[Any]], StructuralResult]]:
-        transformed = [v.bimap(ctx) if isinstance(v, StructuralResult) else (v, lambda x: x) for v in self.value]
-        backmaps = [b for (_, b) in transformed]
-        ret = [a for (a, _) in transformed]
-        def backward(data: List[Any]) -> StructuralResult:
-            if len(data) != len(transformed):
-                raise ValueError("Incompatible data length")
-            return ManyResult(value=tuple([backmaps[i](x) for i, x in enumerate(data)]))
-        return ret, lambda data: backward(data)
-
-
-
-@dataclass(eq=True, frozen=True)
-class OrResult(Generic[A], StructuralResult):
-    value: A
-    def bimap(self, ctx: Any) -> Tuple[Any, Callable[[Any], StructuralResult]]:
-        value, backward = self.value.bimap(ctx) if isinstance(self.value, StructuralResult) else (self.value, lambda x: x)
-        return value, lambda data: OrResult(value=backward(data))
-
-
-class ThenKind(Enum):
-    BOTH = '+'
-    LEFT = '//'
-    RIGHT = '>>'
-    
-@dataclass(eq=True, frozen=True)
-class ThenResult(Generic[A, B], StructuralResult):
-    kind: ThenKind
-    left: A
-    right: B
-    def bimap(self, ctx: Any) -> Tuple[Any, Callable[[Any], StructuralResult]]:
-        def branch(b: Any) -> Tuple[Any, Callable[[Any], StructuralResult]]:
-            return b.bimap(ctx) if isinstance(b, StructuralResult) else (b, lambda x: x)
-        match self.kind:
-            case ThenKind.BOTH:
-                left_value, left_bmap = branch(self.left)
-                right_value, right_bmap = branch(self.right)
-                def backward(x: Tuple[Any, Any]) -> StructuralResult:
-                    return ThenResult(self.kind, left_bmap(x[0]), right_bmap(x[1]))
-                x, y = ThenResult.flat((left_value, right_value))
-                return x, lambda data: backward(y(data))
-            case ThenKind.LEFT:
-                left_value, left_bmap = branch(self.left)
-                return left_value, lambda data: ThenResult(self.kind, left_bmap(data), self.right)
-            case ThenKind.RIGHT:
-                right_value, right_bmap = branch(self.right)
-                return right_value, lambda data: ThenResult(self.kind, self.left, right_bmap(data))
-    @staticmethod
-    def flat(array: Tuple[Any, Any]) -> Tuple[Tuple[Any, ...], Callable[[Tuple[Any, ...]], Tuple[Any, Any]]]:
-        index: Dict[int, int] = {}
-        ret: List[Any] = []
-        for e in array:
-            if isinstance(e, tuple):
-                index[len(ret)] = len(e)
-                ret.extend(e)
-            else:
-                ret.append(e)
-        def backward(data: Tuple[Any, ...]) -> Tuple[Any, Any]:
-            tmp: List[Any] = []
-            skip: int = 0
-            for i, e in enumerate(data):
-                if skip <= 0:
-                    if i in index:
-                        tmp.append(tuple(data[i:i + index[i]]))
-                        skip = index[i] - 1
-                    else:
-                        tmp.append(e)
-                else:
-                    skip -= 1
-            return tuple(tmp)
-        return tuple(ret), backward
+B = TypeVar('B')  # Mapped result type
 
 
 InProgress = object()  # Marker for in-progress state, used to prevent re-entrance in recursive calls
@@ -233,12 +42,13 @@ class Right(Either[L, R]):
 
 
 @dataclass(frozen=True)
-class Error(Insptectable):
+class Error:
     this: Any
     message: Optional[str] = None
     error: Optional[Any] = None    
     state: Optional[Any] = None
     committed: bool = False
+    stack: Optional[str] = None
     previous: Optional[Error] = None
     
     def attach( self, 
@@ -254,59 +64,16 @@ class Error(Insptectable):
             state=state,
             previous=self
         )
-    
+    def to_list(self)->List[Dict[str, Any]]:
+        lst = []
+        current: Optional[Error] = self
+        while current is not None:
+            d = shallow_dict(current)
+            lst.append({k:v for k,v in d.items() if v is not None and k != 'previous'})
+            current = current.previous
+        return lst
 
 
-
-    def to_list(self, interested: Callable[[Any], bool]) -> List[Dict[str, str]]:
-
-        def to_dict() -> Dict[str, str]:
-            data: Dict[str, str] = {}
-            for f in fields(self):
-                value = getattr(self, f.name)
-                if isinstance(value, Error):
-                    # self.previous
-                    pass
-                elif isinstance(value, Insptectable):
-                    # self.this
-                    def inst(x: Any) -> bool:
-                        return x in self.algebras
-                    s = value.to_string(inst) 
-                    data[f.name] = s if s is not None else repr(value)
-                elif value is not None:
-                    # self.committed, self.message, self.expect, self.exception
-                    data[f.name] = repr(value)
-            return data
-        ret = []
-        tmp : None | Error = self
-        while tmp is not None:
-            ret.append(to_dict())
-            tmp = tmp.previous
-        return ret
-
-    def find(self, predicate: Callable[[Error], bool]) -> Optional[Error]:
-        if predicate(self):
-            return self
-        if self.previous is not None:
-            return self.previous.find(predicate)
-        return None
-    
-    def to_string(self, interested: Callable[[Any], bool])->str:
-        lst = self.to_list(interested)
-        root, leaf = lst[0], lst[-1]
-        root_fields = ',\n  '.join([f"{key}: {value}" for key, value in root.items() if value is not None])
-        leaf_fields = ',\n  '.join([f"{key}: {value}" for key, value in leaf.items() if value is not None])
-
-        return  f"{self.__class__.__name__}: ROOT\n"\
-                f"  {root_fields}\n"\
-                f"\u25cf \u25cf \u25cf  LEAF\n"\
-                f"  {leaf_fields}\n"
-
-
-    @cached_property
-    def algebras(self) -> List[Any]:
-        return [self.this] + (self.previous.algebras if self.previous is not None else [])
-        
 @dataclass(frozen=True)        
 class Algebra(ABC, Generic[A, S]):
 ######################################################## shared among all subclasses ########################################################
@@ -349,14 +116,15 @@ class Algebra(ABC, Generic[A, S]):
                     result = Left(result.value.attach(this=self, state=input))
         except Exception as e:
             cache.pop(input, None)  # Clear the cache entry on exception
-            traceback.print_exc()
-            print(f"Exception from self.run(S): {e}")
+            # traceback.print_exc()
+            # print(f"Exception from self.run(S): {e}")
             return Left(
                 Error(
                     message="Exception from self.run(S): {e}",
                     this=self,
                     state=input,
-                    error=e
+                    error=e,
+                    stack=traceback.format_exc()
                 ))
         return result
 
@@ -368,10 +136,7 @@ class Algebra(ABC, Generic[A, S]):
         def lazy_run(input: S, use_cache:bool) -> Either[Any, Tuple[A, S]]:
             return thunk().run(input, use_cache)
         return cls(lazy_run, name=cls.__name__ + '.lazy')
-
-
-
-
+    
     @classmethod
     def fail(cls, error: Any) -> Algebra[Any, S]:
         def fail_run(input: S, use_cache:bool) -> Either[Any, Tuple[Any, S]]:
@@ -381,6 +146,7 @@ class Algebra(ABC, Generic[A, S]):
                 state=input
             ))
         return cls(fail_run, name=cls.__name__ + '.fail')
+    
     @classmethod
     def success(cls, value: Any) -> Algebra[Any, S]:
         def success_run(input: S, use_cache:bool) -> Either[Any, Tuple[Any, S]]:
@@ -474,16 +240,51 @@ class Algebra(ABC, Generic[A, S]):
         lazy_self = self.__class__(debug_run, name=label)  
         return lazy_self
 
+######################################################## map on state ###########################################
+    def post_state(self, f: Callable[[S], S]) -> Algebra[A, S]:
+        def post_state_run(input: S, use_cache:bool) -> Either[Any, Tuple[A, S]]:
+            match self.run(input, use_cache):
+                case Right((value, state)):
+                    return Right((value, f(state)))
+                case Left(err):
+                    return Left(err)
+                case x:
+                    raise ValueError(f"Unexpected result from self.run {x}")
+        return self.__class__(post_state_run, name=self.name) 
 
-######################################################## fundamental combinators ############################################
-    def map(self, f: Callable[[A], B]) -> Algebra[B, S]:
-        def map_run(input: S, use_cache:bool) -> Either[Any, Tuple[B, S]]:
+    def pre_state(self, f: Callable[[S], S]) -> Algebra[A, S]:
+        def pre_state_run(state: S, use_cache:bool) -> Either[Any, Tuple[A, S]]:
+            return self.run(f(state), use_cache)
+        return self.__class__(pre_state_run, name=self.name) 
+
+
+    def map_all(self, f: Callable[[A, S], Tuple[B, S]]) -> Algebra[B, S]:
+        def map_all_run(input: S, use_cache:bool) -> Either[Any, Tuple[B, S]]:
+            match self.run(input, use_cache):
+                case Right((value, state)):
+                    new_value, new_state = f(value, state)
+                    return Right((new_value, new_state))
+                case Left(err):
+                    return Left(err)
+                case x:
+                    raise ValueError(f"Unexpected result from self.run {x}")
+        return self.__class__(map_all_run, name=self.name) # type: ignore
+######################################################## fundamental combinators ############################################    
+    def fmap(self, f: Callable[[A], B]) -> Algebra[B, S]:
+        def fmap_run(input: S, use_cache:bool) -> Either[Any, Tuple[B, S]]:
             parsed = self.run(input, use_cache)
             if isinstance(parsed, Right):
                 return Right((f(parsed.value[0]), parsed.value[1]))            
             else:
                 return cast(Either[Any, Tuple[B, S]], parsed)
-        return self.__class__(map_run, name=self.name)  # type: ignore
+        return self.__class__(fmap_run, name=self.name)  # type: ignore
+
+    
+    def map(self, f: Callable[[A], B]) -> Algebra[B, S]:
+        return self.fmap(f)
+    
+    def bimap(self, f: Callable[[A], B], i: Callable[[B], A]) -> Algebra[B, S]:
+        return self.fmap(f).pre_state(lambda s: s.map(i))
 
     def map_error(self, f: Callable[[Optional[Any]], Any]) -> Algebra[A, S]:
         def map_error_run(input: S, use_cache:bool) -> Either[Any, Tuple[A, S]]:
@@ -493,12 +294,6 @@ class Algebra(ABC, Generic[A, S]):
             return parsed
         return self.__class__(map_error_run, name=self.name)  
 
-    def map_state(self, f: Callable[[S], S]) -> Algebra[A, S]:
-        def map_state_run(state: S, use_cache:bool) -> Either[Any, Tuple[A, S]]:
-            return self.run(f(state), use_cache)
-        return self.__class__(map_state_run, name=self.name) 
-
-
     def flat_map(self, f: Callable[[A], Algebra[B, S]]) -> Algebra[B, S]:
         def flat_map_run(input: S, use_cache:bool) -> Either[Any, Tuple[B, S]]:
             parsed = self.run(input, use_cache)
@@ -509,41 +304,48 @@ class Algebra(ABC, Generic[A, S]):
         return self.__class__(flat_map_run, name=self.name)  # type: ignore
 
     
-    def or_else(self: Algebra[A, S], other: Algebra[B, S]) -> Algebra[OrResult[A | B], S]:
-        def or_else_run(input: S, use_cache:bool) -> Either[Any, Tuple[OrResult[A | B], S]]:
+    def or_else(self: Algebra[A, S], other: Algebra[B, S]) -> Algebra[Choice[A, B], S]:
+        def or_else_run(input: S, use_cache:bool) -> Either[Any, Tuple[Choice[A, B], S]]:
             match self.run(input, use_cache):
                 case Right((value, state)):
-                    return Right((OrResult(value=value), state))
+                    return Right((Choice(kind=ChoiceKind.LEFT, value=value), state))
                 case Left(err):
                     if isinstance(err, Error) and err.committed:
-                        return Left(err)
+                        return Left(replace(err, committed=False))
                     match other.run(input, use_cache):
                         case Right((other_value, other_state)):
-                            return Right((OrResult(value=other_value), other_state))
+                            return Right((Choice(kind=ChoiceKind.RIGHT, value=other_value), other_state))
                         case Left(other_err):
                             return Left(other_err)
                     raise TypeError(f"Unexpected result type from {other}")
             raise TypeError(f"Unexpected result type from {self}")
         return self.__class__(or_else_run, name=f'{self.name} | {other.name}')  # type: ignore
 
-    def then_both(self, other: 'Algebra[B, S]') -> 'Algebra[ThenResult[A, B], S]':
-        def then_both_f(a: A) -> Algebra[ThenResult[A, B], S]:
-            def combine(b: B) -> ThenResult[A, B]:
-                return ThenResult(left=a, right=b, kind=ThenKind.BOTH)
-            return other.map(combine)
+    def then_both(self, other: Algebra[B, S]) -> Algebra[Then[A, B], S]:
+        def then_both_f(a: A) -> Algebra[Then[A, B], S]:
+            def combine(b: B) -> Then[A, B]:
+                return Then(left=a, right=b, kind=ThenKind.BOTH)
+            return other.fmap(combine)
         return self.flat_map(then_both_f).named(f'{self.name} + {other.name}')
-            
-    def then_left(self, other: Algebra[B, S]) -> Algebra[ThenResult[A, B], S]:
-        return self.then_both(other).map(lambda b: replace(b, kind = ThenKind.LEFT)).named(f'{self.name} // {other.name}')
 
-    def then_right(self, other: Algebra[B, S]) -> Algebra[ThenResult[A, B], S]:
-        return self.then_both(other).map(lambda b: replace(b, kind=ThenKind.RIGHT)).named(f'{self.name} >> {other.name}')
+    def then_left(self, other: Algebra[B, S]) -> Algebra[Then[A, B], S]:
+        def then_left_f(a: A) -> Algebra[Then[A, B], S]:
+            def combine(b: B) -> Then[A, B]:
+                return Then(left=a, right=b, kind=ThenKind.LEFT)
+            return other.fmap(combine)
+        return self.flat_map(then_left_f).named(f'{self.name} // {other.name}')
 
+    def then_right(self, other: Algebra[B, S]) -> Algebra[Then[A, B], S]:
+        def then_right_f(a: A) -> Algebra[Then[A, B], S]:
+            def combine(b: B) -> Then[A, B]:
+                return Then(left=a, right=b, kind=ThenKind.RIGHT)
+            return other.fmap(combine)
+        return self.flat_map(then_right_f).named(f'{self.name} >> {other.name}')
 
-    def many(self, *, at_least: int, at_most: Optional[int]) -> Algebra[ManyResult[A], S]:
-        assert at_least > 0, "at_least must be greater than 0"
-        assert at_most is None or at_least <= at_most, "at_least must be less than or equal to at_most"
-        def many_run(input: S, use_cache:bool) -> Either[Any, Tuple[ManyResult[A], S]]:
+    def many(self, *, at_least: int, at_most: Optional[int]) -> Algebra[Many[A], S]:
+        if at_least <=0 or (at_most is not None and at_most < at_least):
+            raise ValueError(f"Invalid arguments for many: at_least={at_least}, at_most={at_most}")
+        def many_run(input: S, use_cache:bool) -> Either[Any, Tuple[Many[A], S]]:
             ret: List[A] = []
             current_input = input
             while True:
@@ -567,7 +369,7 @@ class Algebra(ABC, Generic[A, S]):
                         this=self,
                         state=current_input
                     )) 
-            return Right((ManyResult(value=tuple(ret)), current_input))
+            return Right((Many(value=tuple(ret)), current_input))
         return self.__class__(many_run, name=f'*({self.name})') # type: ignore
 
     
