@@ -2,31 +2,27 @@ from __future__ import annotations
 import re
 from sqlglot import tokenize, TokenType, Parser as GlotParser, exp
 from typing import (
-    Optional, List, Any, Tuple,
-    Generic, Callable
+    Optional, List, Any, Tuple, TypeVar,
+    Generic
 )
 from syncraft.algebra import (
-    Either, Left, Right, Error, Insptectable, Algebra
+    Either, Left, Right, Error, Algebra
 )
 from dataclasses import dataclass, field, replace
 from enum import Enum
 from functools import reduce
-from syncraft.dsl import DSL
-
-from syncraft.ast import Token, TokenSpec, AST, T
-
-
-
-
-
-
+from syncraft.syntax import Syntax
 
+from syncraft.ast import Token, TokenSpec, AST, TokenProtocol
+from syncraft.constraint import Bindable
 
 
+T = TypeVar('T', bound=TokenProtocol)
 @dataclass(frozen=True)
-class ParserState(Generic[T], Insptectable):
+class ParserState(Bindable, Generic[T]):
     input: Tuple[T, ...] = field(default_factory=tuple)
     index: int = 0
+
     
     def token_sample_string(self)-> str:
         def encode_tokens(*tokens:T) -> str:
@@ -40,13 +36,6 @@ class ParserState(Generic[T], Insptectable):
     def after(self, length: Optional[int] = 5)->str:
         length = min(length, len(self.input) - self.index) if length is not None else len(self.input) - self.index
         return " ".join(token.text for token in self.input[self.index:self.index + length])
- 
-    def to_string(self, interested: Callable[[Any], bool])->str:
-        return f"ParserState(\n"\
-               f"index={self.index}, \n"\
-               f"input({len(self.input)})=[{self.token_sample_string()}, ...]), \n"\
-               f"before=({self.before()}), \n"\
-               f"after=({self.after()})"  
 
 
     def current(self)->T:
@@ -187,19 +176,20 @@ class Parser(Algebra[T, ParserState[T]]):
             return Right((tuple(tokens), tmp_state))
         return cls(until_run, name=cls.__name__ + '.until')
 
-def sqlglot(parser: DSL[Any, Any], 
-            dialect: str) -> DSL[List[exp.Expression], ParserState[Any]]:
+def sqlglot(parser: Syntax[Any, Any], 
+            dialect: str) -> Syntax[List[exp.Expression], ParserState[Any]]:
     gp = GlotParser(dialect=dialect)
     return parser.map(lambda tokens: [e for e in gp.parse(raw_tokens=tokens) if e is not None])
 
 
-def parse(parser: Algebra[Any, ParserState[Token]], 
+def parse(syntax: Syntax[Any, Any], 
           sql: str, 
-          dialect: str) -> AST[Any] | Any:
+          dialect: str) -> AST | Any:
+    parser = syntax(Parser)
     input: ParserState[Token] = token_state(sql, dialect=dialect)
     result = parser.run(input, True)
     if isinstance(result, Right):
-        return AST(result.value[0])
+        return result.value[0]
     assert isinstance(result, Left), "Parser must return Either[E, Tuple[A, S]]"
     return result.value
 
@@ -212,34 +202,34 @@ def token(token_type: Optional[Enum] = None,
           text: Optional[str] = None, 
           case_sensitive: bool = False,
           regex: Optional[re.Pattern[str]] = None
-          ) -> DSL[Any, Any]:
+          ) -> Syntax[Any, Any]:
     token_type_txt = token_type.name if token_type is not None else None
     token_value_txt = text if text is not None else None
     msg = 'token(' + ','.join([x for x in [token_type_txt, token_value_txt, str(regex)] if x is not None]) + ')'
-    return DSL(
+    return Syntax(
         lambda cls: cls.factory('token', token_type=token_type, text=text, case_sensitive=case_sensitive, regex=regex)
         ).describe(name=msg, fixity='prefix') 
 
     
-def identifier(value: str | None = None) -> DSL[Any, Any]:
+def identifier(value: str | None = None) -> Syntax[Any, Any]:
     if value is None:
         return token(TokenType.IDENTIFIER)
     else:
         return token(TokenType.IDENTIFIER, text=value)
 
-def variable(value: str | None = None) -> DSL[Any, Any]:
+def variable(value: str | None = None) -> Syntax[Any, Any]:
     if value is None:
         return token(TokenType.VAR)
     else:
         return token(TokenType.VAR, text=value)
 
-def literal(lit: str) -> DSL[Any, Any]:
+def literal(lit: str) -> Syntax[Any, Any]:
     return token(token_type=None, text=lit, case_sensitive=True)
 
-def regex(regex: re.Pattern[str]) -> DSL[Any, Any]:
+def regex(regex: re.Pattern[str]) -> Syntax[Any, Any]:
     return token(token_type=None, regex=regex, case_sensitive=True)
 
-def lift(value: Any)-> DSL[Any, Any]:
+def lift(value: Any)-> Syntax[Any, Any]:
     if isinstance(value, str):
         return literal(value)
     elif isinstance(value, re.Pattern):
@@ -247,22 +237,22 @@ def lift(value: Any)-> DSL[Any, Any]:
     elif isinstance(value, Enum):
         return token(value)
     else:
-        return DSL(lambda cls: cls.success(value))
+        return Syntax(lambda cls: cls.success(value))
 
-def number() -> DSL[Any, Any]:
+def number() -> Syntax[Any, Any]:
     return token(TokenType.NUMBER)
 
 
-def string() -> DSL[Any, Any]:
+def string() -> Syntax[Any, Any]:
     return token(TokenType.STRING)
 
 
 
-def until(*open_close: Tuple[DSL[Tuple[T, ...] | T, ParserState[T]], DSL[Tuple[T, ...] | T, ParserState[T]]],
-          terminator: Optional[DSL[Tuple[T, ...] | T, ParserState[T]]] = None,
+def until(*open_close: Tuple[Syntax[Tuple[T, ...] | T, ParserState[T]], Syntax[Tuple[T, ...] | T, ParserState[T]]],
+          terminator: Optional[Syntax[Tuple[T, ...] | T, ParserState[T]]] = None,
           inclusive: bool = True, 
-          strict: bool = True) -> DSL[Any, Any]:
-    return DSL(
+          strict: bool = True) -> Syntax[Any, Any]:
+    return Syntax(
         lambda cls: cls.factory('until', 
                            *[(left.alg(cls), right.alg(cls)) for left, right in open_close], 
                            terminator=terminator.alg(cls) if terminator else None, 
