@@ -1,4 +1,5 @@
-from syncraft.parser import AST, literal, variable, parse, Parser
+from syncraft.ast import AST
+from syncraft.parser import literal, variable, parse
 import syncraft.generator as gen
 from typing import Any
 
@@ -12,17 +13,21 @@ var = variable()
 def test_between()->None:
     sql = "then if then"
     syntax = IF.between(THEN, THEN)
-    ast:AST[Any] = parse(syntax(Parser), sql, dialect='sqlite')    
-    generated = gen.generate(syntax(gen.Generator), ast)
+    ast:AST = parse(syntax, sql, dialect='sqlite')    
+    generated = gen.generate(syntax, ast)
     assert ast == generated, "Parsed and generated results do not match."
+    x, f = generated.bimap()
+    assert gen.generate(syntax, f(x)) == ast
 
 
 def test_sep_by()->None:
     sql = "if then if then if then if"
     syntax = IF.sep_by(THEN)
-    ast:AST[Any] = parse(syntax(Parser), sql, dialect='sqlite')    
-    generated = gen.generate(syntax(gen.Generator), ast)
+    ast:AST = parse(syntax, sql, dialect='sqlite')    
+    generated = gen.generate(syntax, ast)
     assert ast == generated, "Parsed and generated results do not match."
+    x, f = generated.bimap()
+    assert gen.generate(syntax, f(x)) == ast
 
 def test_many_or()->None:
     IF = literal("if")
@@ -30,6 +35,8 @@ def test_many_or()->None:
     END = literal("end")
     syntax = (IF.many() | THEN.many()).many() // END
     sql = "if if then end"
-    ast:AST[Any] = parse(syntax(Parser), sql, dialect='sqlite')
-    generated = gen.generate(syntax(gen.Generator), ast)
+    ast:AST = parse(syntax, sql, dialect='sqlite')
+    generated = gen.generate(syntax, ast)
     assert ast == generated, "Parsed and generated results do not match."
+    x, f = generated.bimap()
+    assert gen.generate(syntax, f(x)) == ast
